import logging
from inspect import isawaitable

from sanic.response import json

from brama.plugins import MiddlewarePlugin, AuthorizationPlugin

logger = logging.getLogger(__name__)


async def get_auth_value(request, plugin):
    try:
        result = plugin.get_authorization_value(request)
        if isawaitable(result):
            result = await result
        return result
    except NotImplementedError:
        pass

    # Take from Authorization header by default
    header_name = request.app.config['BRAMA_AUTHORIZATION_HEADER']
    return request.headers.get(header_name.lower())


@MiddlewarePlugin.register()
async def authorization(request):
    request.authorization = None

    _config = request.route.get_config('authorization')
    if not _config:
        # No authorization configured for this endpoint.
        return

    if not isinstance(_config, list):
        auth_plugins = [_config]
    else:
        auth_plugins = _config

    for plugin_ in auth_plugins:
        if isinstance(plugin_, str):
            plugin = request.app.plugins_registry.get(plugin_)
        else:
            plugin = plugin_

        if plugin_ and plugin is None:
            raise ValueError('Can not find plugin {}'.format(plugin_))

        if not isinstance(plugin, AuthorizationPlugin):
            err = 'Expect <AuthorizationPlugin>, got {} instead'.format(plugin)
            raise TypeError(err)

        auth_value = await get_auth_value(request, plugin)
        try:
            authorization = plugin.process_authorization_value(request.app,
                                                               auth_value)
            if isawaitable(authorization):
                authorization = await authorization

        except NotImplementedError:
            authorization = None

        if authorization is not None:
            request.authorization = authorization
            break

    if not request.authorization:
        return json({}, status=401)



import argparse

from .app import Brama

parser = argparse.ArgumentParser(description='Brama API Gateway')
parser.add_argument('--config', '-c', dest='config', help='Path to config',
                    required=True)
parser.add_argument('--plugins', dest='plugins',
                    help='Path to plugins python modules')
parser.add_argument('--host', '-s', dest='host', help='Host',
                    default='0.0.0.0')
parser.add_argument('--port', '-p', dest='port', help='Port', default=8051,
                    type=int)
parser.add_argument('--workers', '-w', dest='workers', help='Workers num',
                    default=1, type=int)
parser.add_argument('--debug', dest='debug', help='Debug', action='store_true')

args = parser.parse_args()

brama = Brama('brama')
brama.load_config(args.config)
brama.load_plugins(args.plugins)


def main():
    brama.run(args.host, args.port, debug=args.debug, workers=args.workers)


if __name__ == '__main__':
    main()



class API(object):
    name = None
    verbose_name = None
    active = True
    routes = []
    plugins = []
    _extra_config = {}

    def __init__(self, name=None, verbose_name=None, routes=None, plugins=None,
                 active=None, **kwargs):
        if name:
            self.name = name
        if verbose_name:
            self.verbose_name = verbose_name
        if routes is not None:
            self.routes = routes
        if plugins is not None:
            self.plugins = plugins
        if active is not None:
            self.active = active

        self._extra_config = {}
        for k, v in kwargs.items():
            self._extra_config[k] = v

    def get_config(self, param_name):
        return self._extra_config.get(param_name, None)



from .utils import Registry


class ApiRegistry(Registry):

    def add(self, api_name, api):
        self.add_item(api_name, api)

    def get(self, api_name):
        return self.get_item(api_name)


class PluginsRegistry(Registry):

    def add(self, plugin):
        self.add_item(plugin.name, plugin)

    def get(self, name, type_=None):
        plugin = self.get_item(name)
        if plugin and type_ and plugin.type != type_:
            raise TypeError('Plugin with name {} is not of type {}'.format(
                name, type_
            ))
        return plugin


api_registry = ApiRegistry()
plugins_registry = PluginsRegistry()



import random
from inspect import isawaitable
from urllib.parse import urljoin

import aiohttp
from sanic.request import Request
from sanic.response import HTTPResponse, StreamingHTTPResponse

from brama.plugins import RouteHandlerPlugin
from .request import ProxyRequest, LocalRequest


class BaseHandler(object):
    request_class = None
    is_stream = False  # for Sanic, allows to use request.stream

    def __init__(self, route, prefix_path=None):
        self.route = route
        self.prefix_path = prefix_path
        self.is_stream = self.route.stream

    async def __call__(self, request: Request, *view_args, **view_kwargs):
        return await self._process_request(request, view_args, view_kwargs)

    async def _process_request(self, request: Request, view_args, view_kwargs):
        api_request = self.api_request_factory(request, view_args, view_kwargs)
        response = await self.run_request_middlewares(api_request)
        if response is None:
            response = await self.get_response(api_request)

        _response = await self.run_response_middlewares(api_request, response)
        return _response if _response is not None else response

    def api_request_factory(self, request: Request, view_args, view_kwargs):
        return self.request_class(self.route, request,
                                  prefix_path=self.prefix_path,
                                  view_args=view_args,
                                  view_kwargs=view_kwargs)

    async def run_request_middlewares(self, api_request):
        for plugin in api_request.enabled_plugins(
                required_methods=['process_request']):
            resp = plugin.process_request(api_request)
            if isawaitable(resp):
                resp = await resp

            if resp:
                return resp
        return None

    async def run_response_middlewares(self, api_request, response):
        for plugin in api_request.enabled_plugins(
                required_methods=['process_response']):
            resp = plugin.process_response(api_request, response)
            if isawaitable(resp):
                resp = await resp

            if resp:
                return resp
        return None

    async def get_response(self, api_request):
        raise NotImplementedError()


class ProxyHandler(BaseHandler):
    request_class = ProxyRequest

    async def _proxy_normal_response(self, api_request, url):
        async with aiohttp.ClientSession(auto_decompress=False) as sess:
            async with sess.request(api_request.method, url,
                                    data=api_request.proxy_data,
                                    headers=api_request.headers) as r:
                return HTTPResponse(body_bytes=await r.content.read(),
                                    status=r.status,
                                    headers=dict(r.headers))

    async def _proxy_stream_response(self, api_request, url):
        # TODO: Refactor to make sure session always closed.
        # and maybe add keep-alive support with upstreams

        client_session = aiohttp.ClientSession(auto_decompress=False)
        req = client_session.request(api_request.method, url,
                                     data=api_request.proxy_data,
                                     headers=api_request.headers)
        resp = await req._coro

        async def streaming_fn(response):
            try:
                async for data in resp.content.iter_any():
                    response.write(data)
            finally:
                # TODO: Websockets needs `resp.close()`
                await resp.release()
                await client_session.close()

        return StreamingHTTPResponse(streaming_fn,
                                     status=resp.status,
                                     headers=dict(resp.headers))

    def get_upstream_url(self, api_request):
        targets = self.route.upstreams.get('targets', [])
        strategy = self.route.upstreams.get('balancing', 'random')

        if strategy == 'roundrobin':
            raise NotImplementedError()
        elif strategy == 'random':
            upstream = random.choice(targets)
        else:
            raise NotImplementedError()

        # Fill upstream url 'variables' with parsed route's values.
        # `url_vars` can be extended with another helpful params, like
        # request_id, user_id, api_version, etc.
        url_vars = dict(api_request.view_kwargs)
        upstream = upstream.format(**url_vars)

        url = urljoin(upstream, api_request.path.lstrip('/'))
        url += api_request.query_string
        return url

    async def get_response(self, api_request):
        url = self.get_upstream_url(api_request)
        try:
            if api_request.route.stream:
                response = await self._proxy_stream_response(api_request, url)
            else:
                response = await self._proxy_normal_response(api_request, url)
        except aiohttp.client_exceptions.ClientConnectorError:
            # TODO
            response = HTTPResponse('Bad Gateway', status=502)
        return response


class LocalHandler(BaseHandler):
    request_class = LocalRequest

    def _get_request_handler(self, api_request):
        '''
        Returns view object basing on specified route's handler.
        :return: function
        '''
        if isinstance(self.route.handler, str):
            plugin = api_request.app.plugins_registry.get(self.route.handler,
                                                          'route_handler')
            if not plugin:
                raise ValueError('Handler {} could not be found'.format(
                    self.route.handler))
            handler_func = plugin.wrapped_object
        elif isinstance(self.route.handler, RouteHandlerPlugin):
            handler_func = self.route.handler.wrapped_object
        else:
            handler_func = self.route.handler
        return handler_func

    async def get_response(self, api_request):
        # TODO: add support for classview in future
        func = self._get_request_handler(api_request)
        result = func(api_request._original_request,
                      *api_request.view_args,
                      **api_request.view_kwargs)

        if isawaitable(result):
            result = await result

        if isinstance(result, HTTPResponse):
            return result

        status = 200
        if isinstance(result, tuple):
            data, status = result
        else:
            data = result

        return HTTPResponse(data, status=status)



from collections import defaultdict
from threading import RLock


def pathjoin(*parts):
    all_parts = []
    traling_slash = False
    if parts and parts[-1].endswith('/'):
        traling_slash = True
    for part in parts:
        all_parts.extend(part.split('/'))
    joined = '/'.join([p for p in all_parts if p])
    if traling_slash:
        joined += '/'
    return '/%s' % joined


def yaml_config_parse(conf_fp):
    try:
        import yaml
    except ImportError:
        raise EnvironmentError(
            'To use YAML configuration please install PyYAML or '
            'install brama as `pip install brama[yaml]`'
        )
    return yaml.load(conf_fp)


def unique_items(iterable, attr_name='name'):
    '''
    Returns list without duplicated plugins/apis. Works with them as strings
    (names) and instances.
    :param iterable:
    :param attr_name: for objects in iterable, use attr's `attr_name` value
    :return: generator
    '''
    unique_set = set()
    for item in iterable:
        if isinstance(item, str):
            value = item
        else:
            value = getattr(item, attr_name)
        if value in unique_set:
            continue
        unique_set.add(value)
        yield item


class Registry(object):
    _lock = RLock()
    _items = None

    def __init__(self, app=None):
        if app:
            self.init_app(app)
        self._items = defaultdict(dict)

    def init_app(self, app):
        self.app = app

    def add_item(self, name, value):
        if name is None:
            raise ValueError('Name is required to be registered in registry')
        if value is None:
            raise ValueError('Value of `{}` can\'t be null'.format(name))

        with self._lock:
            if name in self._items:
                raise ValueError('Already in registry: {}'.format(name))
            self._items[name] = value
        return

    def get_item(self, name):
        with self._lock:
            return self._items.get(name)

    def items(self):
        return self._items.values()



from . import registry


class Plugin(object):
    type = None
    _required_config = []
    _wrapped_object = None

    def __init__(self, name, required_config=None):
        self.name = name
        if required_config:
            self._required_config = required_config

    @classmethod
    def register(cls, plugin_name=None, *args, **kwargs):
        def wrapper(obj):
            name = plugin_name if plugin_name else obj.__name__
            plugin = cls.register_new(name, *args, **kwargs)
            plugin.wrapped_object = obj
            plugin.required_config = kwargs.get('required_config', [])
            registry.plugins_registry.add(plugin)
            return obj

        return wrapper

    @classmethod
    def register_new(cls, name, *args, **kwargs):
        return cls(name, *args, **kwargs)

    @property
    def wrapped_object(self):
        return self._wrapped_object

    @wrapped_object.setter
    def wrapped_object(self, obj):
        self._wrapped_object = obj


class RouteHandlerPlugin(Plugin):
    type = 'route_handler'


class RoutesPlugin(Plugin):
    type = 'routes'

    def get_extra_routes(self, app):
        return self.wrapped_object(app) or []


class AuthorizationFlowPlugin(Plugin):
    type = 'authorization_flow'

    def process_login(self, request, username, password):
        obj = self.wrapped_object()
        if not hasattr(obj, 'process_login'):
            raise NotImplementedError()
        return obj.process_login(request, username, password)

    def get_extra_routes(self, app):
        '''
        Allow Authorization plugins to define routes.
        :return: list
        '''
        obj = self.wrapped_object()
        if not hasattr(obj, 'get_extra_routes'):
            return []
        return obj.get_extra_routes(app)


class AuthorizationPlugin(Plugin):
    type = 'authorization'

    def get_authorization_value(self, request):
        obj = self.wrapped_object()
        if not hasattr(obj, 'get_authorization_value'):
            raise NotImplementedError()
        return obj.get_authorization_value(request)

    def process_authorization_value(self, app, value):
        obj = self.wrapped_object()
        if not hasattr(obj, 'process_authorization_value'):
            raise NotImplementedError()
        return obj.process_authorization_value(app, value)


class MiddlewarePlugin(Plugin):
    type = 'middleware'
    attached_to = 'request'

    def __init__(self, name, attached_to=None):
        if attached_to is not None:
            self.attached_to = attached_to
        super(MiddlewarePlugin, self).__init__(name)

    @classmethod
    def register_new(cls, name, *args, **kwargs):
        attached_to = kwargs.get('attached_to', 'request')
        return cls(name, attached_to)

    def process_request(self, *args, **kwargs):
        if self.attached_to == 'request':
            return self.wrapped_object(*args, **kwargs)

    def process_response(self, *args, **kwargs):
        if self.attached_to == 'response':
            return self.wrapped_object(*args, **kwargs)


class RequestTransformPlugin(MiddlewarePlugin):
    type = 'request_transformation'
    attached_to = 'request'

    @classmethod
    def register_new(cls, name, *args, **kwargs):
        return cls(name)


class ResponseTransformPlugin(MiddlewarePlugin):
    type = 'response_transformation'
    attached_to = 'response'

    @classmethod
    def register_new(cls, name, *args, **kwargs):
        return cls(name)



from aiohttp import streamer
from sanic.request import Request


class ApiRequest(object):
    path = None
    method = None
    headers = None
    query_string = None
    parsed_args = None
    view_args = None
    view_kwargs = None

    def __init__(self, route, original_request: Request, prefix_path=None,
                 view_args=None, view_kwargs=None):
        self.route = route
        self.prefix_path = prefix_path
        self._original_request = original_request
        self.app = original_request.app
        self.path = original_request.path
        self.method = original_request.method
        self.headers = dict(original_request.headers)
        self.parsed_args = original_request.parsed_args
        self.query_string = original_request.query_string
        self.view_args = view_args or tuple()
        self.view_kwargs = view_kwargs or dict()
        self.body_stream = self._get_body_streamer()

    def enabled_plugins(self, required_methods=()):
        return self.app.enabled_plugins(
            api=self.route.api, route=self.route,
            required_methods=required_methods)

    def _get_body_streamer(self):
        @streamer
        def request_body_stream(writer):
            while True:
                chunk = yield from self._original_request.stream.get()
                yield from writer.write(chunk)
                if not chunk:
                    break

        return request_body_stream


class ProxyRequest(ApiRequest):

    def __init__(self, *args, **kwargs):
        super(ProxyRequest, self).__init__(*args, **kwargs)
        self.path = self.get_upstream_path()

    @property
    def proxy_data(self):
        if self._original_request.stream:
            return self.body_stream
        return self._original_request.body

    def get_upstream_path(self):
        upstream_path = self.path[:]

        if self.route.strip_path:
            upstream_path = '/'
        elif not self.route.strip_path:
            if self.prefix_path:
                _clean = self.path[len(self.prefix_path):].lstrip('/')
                upstream_path = '/' + _clean
        return upstream_path

    def replace_path(self, new_path):
        self.path = new_path

    def replace_header(self, name, new_value):
        self.headers[name] = new_value

    def replace_method(self, new_method):
        self.method = new_method


class LocalRequest(ApiRequest):
    pass

    # @property
    # async def json(self):
    #     if not self._original_request.body:
    #         a = self._original_request.stream.read()
    #         raise Exception(a)
    #         self._original_request.body = await self.body_stream.get()
    #     return self._original_request.json



import importlib
import ujson

from sanic import Sanic

from . import default_config
from . import registry
from .api import API
from .route import Route
from .utils import pathjoin, unique_items, yaml_config_parse

BRAMA_PREFIX = 'BRAMA_'


class Brama(Sanic):
    def __init__(self, *args, **kwargs):
        self.api_registry = kwargs.pop('api_registry',
                                       registry.api_registry)
        self.plugins_registry = kwargs.pop('plugins_registry',
                                           registry.plugins_registry)

        if self.api_registry:
            self.api_registry.init_app(self)

        if self.plugins_registry:
            self.plugins_registry.init_app(self)

        self.api_base_path = kwargs.pop('api_base_path', '/')
        self.api_route_class = kwargs.pop('api_route_class', Route)
        self.global_plugins = kwargs.pop('global_plugins', [])

        super(Brama, self).__init__(*args, **kwargs)

        self.config.from_object(default_config)
        self.config.load_environment_vars(BRAMA_PREFIX)

    def register_route(self, route_, api=None, namespace_path=None):
        if isinstance(route_, dict):
            route = self.api_route_class(api, **route_)
        else:
            route = route_

        if namespace_path:
            prefix_path = pathjoin(self.api_base_path, namespace_path)
        else:
            prefix_path = self.api_base_path
        listen_path = pathjoin(prefix_path, route.url_path)

        self.route(
            uri=listen_path,
            methods=route.methods,
            stream=route.stream,
            name=route.name)(route.route_handler_factory(prefix_path))
        return route_

    def register_api(self, obj_or_class):
        if isinstance(obj_or_class, type):
            api = obj_or_class()
        else:
            api = obj_or_class

        if api.active:
            # Deactivated API basically means it's unreachable because not
            # registered in router.
            for route_ in api.routes:
                self.register_route(route_, api=api, namespace_path=api.name)

        self.api_registry.add(api.name, api)
        return obj_or_class

    def load_plugins(self, path_=None):
        path_ = path_ or ''
        pathes = path_.split(',')
        pathes.extend(self.config.get('PLUGINS_MODULES', []))
        pathes.insert(0, 'brama.contrib.plugins')

        for path in [x for x in pathes if x]:
            importlib.import_module(path)

        # Use plugins that provides routes
        for route_plugin in self.enabled_plugins(
                required_methods=['get_extra_routes']):
            [self.register_route(r)
             for r in route_plugin.get_extra_routes(self)]

    def load_config(self, path_or_dict):
        '''
        Known top-level keys of config object are parsed and added to the app.
        Unknown keys are ignored.
        :param path_or_dict: Path to config file or dict
        '''
        if isinstance(path_or_dict, str):
            path = path_or_dict
            if path.endswith('.json'):
                with open(path, 'r') as config_file:
                    config = ujson.load(config_file)
            elif path.endswith('.yaml') or path.endswith('.yml'):
                with open(path, 'r') as config_file:
                    config = yaml_config_parse(config_file)
            else:
                raise ValueError("Unknown config file format")
        else:
            config = dict(path_or_dict)

        self.config.update(config.get('settings', {}))
        self.global_plugins.extend(config.get('global_plugins', []))

        for api_name, api_conf in config.get('apis', []).items():
            api = API(name=api_name, **api_conf)
            self.register_api(api)

    def enabled_plugins(self, api=None, route=None, required_methods=(),
                        types=None):
        '''
        Returns plugins iterator with global enabled plugins, plus
        api and route specific, if those parameters provided.
        :param api: consider enabled if listed in provided api
        :param route: consider enabled if listed in provided route
        :param required_methods: exclude plugins that don't have this methods
        :param types: exclude plugins that has different type
        :return: iterator
        '''
        required_methods = set(required_methods)
        enabled_plugins = list()

        # Form a set if plugins with respect to priority
        enabled_plugins.extend(self.global_plugins)
        if route:
            enabled_plugins.extend(route.plugins)
        if api is not None:
            enabled_plugins.extend(api.plugins)

        for plugin_ in unique_items(enabled_plugins):
            if not isinstance(plugin_, str):
                plugin = plugin_
            else:
                plugin = self.plugins_registry.get(plugin_)
                if not plugin:
                    raise ValueError(
                        'Plugin {} can not be loaded'.format(plugin_))

            if types and plugin.type not in types:
                continue

            # Help plugins to let user know which config vars are required
            # and missing.
            _conf_err = (
                '{} config is required for plugin {}. '
                'You can set it via environment variable named {}'
            )
            if plugin.required_config:
                for _var_name in plugin.required_config:
                    _var_name = _var_name.upper()

                    if _var_name not in self.config.keys():
                        err = _conf_err.format(
                            _var_name, plugin.name,
                            "%s%s" % (BRAMA_PREFIX, _var_name)
                        )
                        raise ValueError(err)

            if required_methods.issubset(set(dir(plugin))):
                yield plugin



import logging
from inspect import isawaitable

import jwt
from sanic.response import json, BaseHTTPResponse

from brama.plugins import AuthorizationFlowPlugin, AuthorizationPlugin
from brama.utils import pathjoin

logger = logging.getLogger(__name__)


@AuthorizationPlugin.register()
class JWTToken(object):
    HEADER_PREFIX = None

    @classmethod
    def _auth_type(cls, app):
        return cls.HEADER_PREFIX or app.config.get('BRAMA_JWT_HEADER_PREFIX')

    @classmethod
    def process_authorization_value(cls, app, value):
        try:
            auth_type, auth_value = value.split(None, 1)
        except (ValueError, AttributeError):
            return
        if auth_type.upper() != cls._auth_type(app).upper():
            return

        payload = cls._decode_token(auth_value, app.config)
        return payload

    @classmethod
    def _decode_token(cls, token, config):
        try:
            payload = jwt.decode(token, config.BRAMA_SECRET_KEY,
                                 algorithm=config.BRAMA_JWT_ALGORITHM)
        except (jwt.exceptions.DecodeError):
            return

        return payload


@AuthorizationFlowPlugin.register()
class jwt_auth(object):

    def process_login(self, request, username, password):
        raise NotImplementedError()

    async def login_handler(self, request):
        app = request.app
        config = app.config
        data = request.json
        try:
            username, password = data['username'], data['password']
        except KeyError:
            return json({'error': 'Invalid data'}, status=400)

        login_plugins = app.enabled_plugins(required_methods=['process_login'])
        result = None
        for plugin in login_plugins:
            try:
                result = plugin.process_login(request, username, password)
                if isawaitable(result):
                    result = await result
            except NotImplementedError:
                continue
            if result:
                break

        if isinstance(result, BaseHTTPResponse):
            return result

        if not result:
            return json({'error': 'Can not authenticate'}, status=401)

        # TODO: https://goo.gl/jpYvwk
        payload = {'some': 'payload'}
        payload.update(result)
        token = jwt.encode(payload, config.BRAMA_SECRET_KEY,
                           algorithm=config.BRAMA_JWT_ALGORITHM)
        return json({'access_token': token})

    def verify_token_handler(self, request):
        config = request.app.config
        token = request.original_request.json.get('access_token')
        if not token:
            return json({'error': 'Invalid params'}, status=400)
        return json({
            'is_valid': bool(self._decode_token(token, config)),
        })

    def revoke_token_handler(self, request):
        return

    def get_extra_routes(self, app):
        path_prefix = app.config.BRAMA_AUTH_PATH
        return [
            {
                'url_path': pathjoin(path_prefix, 'login'),
                'name': 'JWTLogin',
                'handler': self.login_handler,
                'methods': ['POST'],
                'stream': False,
                'authorization': [],
            },
            {
                'url_path': pathjoin(path_prefix, 'verify_token'),
                'name': 'JWTVerify',
                'handler': self.verify_token_handler,
                'methods': ['POST'],
                'stream': False,
                'authorization': [],
            },
            {
                'url_path': pathjoin(path_prefix, 'revoke_token'),
                'name': 'JWTRevoke',
                'handler': self.revoke_token_handler,
                'methods': ['POST'],
                'stream': False,
                'authorization': [],
            },
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import re
from setuptools import setup, find_packages

with io.open('README.rst', 'rt', encoding='utf8') as f:
    readme = f.read()

with io.open('brama/__init__.py', 'rt', encoding='utf8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setup(
    name='brama',
    version=version,
    python_requires='>={}.{}'.format(3, 5),
    url='https://github.com/kolyaflash/brama',
    license='MIT',
    author='Nikolay Baluk',
    author_email='kolyaflash@gmail.com',
    description='API Gateway proxy server with plugins',
    long_description=readme,
    packages=find_packages(exclude=['tests', 'docs', 'venv', 'examples']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'sanic>=0.7',
        'aiohttp>=3.0',
        'PyJWT',
    ],
    extras_require={
        'dev': [
            'pytest>=3',
            'pytest-localserver',
            'coverage',
            'tox',
            'sphinx',
            'flake8',
        ],
        'yaml': [
            'PyYAML>=3.12',
        ]
    },
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Web Environment',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
    ],
    entry_points={
        'console_scripts': [
            'brama = brama.cli:main',
        ],
    },
)



import logging

from brama.plugins import MiddlewarePlugin

logger = logging.getLogger(__name__)


@MiddlewarePlugin.register()
def rate_limiting(request):
    # rate_limits = request.route.get_config('rate_limits')
    return



from .handlers import ProxyHandler, LocalHandler


class Route(object):
    def __init__(self, api=None, name=None, url_path=None, strip_path=False,
                 upstreams=None, handler=None, methods=None, plugins=None,
                 stream=False, **kwargs):
        self.api = api
        self.url_path = url_path
        self.name = name
        self.strip_path = strip_path
        self.upstreams = upstreams
        self.handler = handler
        self.stream = stream
        self.methods = methods if methods else {'GET'}
        self.plugins = plugins if plugins else []

        self._extra_config = {}

        for k, v in kwargs.items():
            self._extra_config[k] = v

        if not self.url_path:
            raise ValueError('url_path is required')

        if self.upstreams and self.handler:
            raise ValueError('Only one of upstreams or handler could be used')
        if not self.upstreams and not self.handler:
            raise ValueError('One of upstreams or handler should be set')

    @property
    def is_proxy(self):
        return True if self.upstreams else False

    def route_handler_factory(self, prefix_path):
        if self.is_proxy:
            kls = ProxyHandler
        else:
            kls = LocalHandler
        return kls(self, prefix_path)

    def get_config(self, param_name):
        route_value = self._extra_config.get(param_name, None)
        api_value = self.api.get_config(param_name) if self.api else None

        if isinstance(route_value, str) or isinstance(api_value,
                                                      str):
            return route_value or api_value

        if isinstance(route_value, list) or isinstance(api_value, list):
            return route_value or api_value

        if isinstance(route_value, dict) or isinstance(api_value, dict):
            _res_dict = {}
            _res_dict.update(api_value) if api_value is not None else None
            _res_dict.update(route_value) if route_value is not None else None
            return _res_dict
